@@ -22,19 +22,31 @@ import (
 	"context"
 	"errors"
 	"fmt"
-	"reflect"
 	"testing"
 
-	"github.com/golang/protobuf/proto"
-	"github.com/golang/protobuf/ptypes"
-	apb "github.com/golang/protobuf/ptypes/any"
-	dpb "github.com/golang/protobuf/ptypes/duration"
+	"github.com/google/go-cmp/cmp"
 	cpb "google.golang.org/genproto/googleapis/rpc/code"
 	epb "google.golang.org/genproto/googleapis/rpc/errdetails"
 	spb "google.golang.org/genproto/googleapis/rpc/status"
+	"google.golang.org/protobuf/proto"
+	"google.golang.org/protobuf/protoadapt"
+	"google.golang.org/protobuf/reflect/protoreflect"
+	"google.golang.org/protobuf/types/known/anypb"
+	"google.golang.org/protobuf/types/known/durationpb"
+
 	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/internal/grpctest"
+	"google.golang.org/grpc/internal/status"
 )
 
+type s struct {
+	grpctest.Tester
+}
+
+func Test(t *testing.T) {
+	grpctest.RunSubTests(t, s{})
+}
+
 // errEqual is essentially a copy of testutils.StatusErrEqual(), to avoid a
 // cyclic dependency.
 func errEqual(err1, err2 error) bool {
@@ -49,20 +61,20 @@ func errEqual(err1, err2 error) bool {
 	return proto.Equal(status1.Proto(), status2.Proto())
 }
 
-func TestErrorsWithSameParameters(t *testing.T) {
+func (s) TestErrorsWithSameParameters(t *testing.T) {
 	const description = "some description"
 	e1 := Errorf(codes.AlreadyExists, description)
 	e2 := Errorf(codes.AlreadyExists, description)
 	if e1 == e2 || !errEqual(e1, e2) {
-		t.Fatalf("Errors should be equivalent but unique - e1: %v, %v  e2: %p, %v", e1.(*statusError), e1, e2.(*statusError), e2)
+		t.Fatalf("Errors should be equivalent but unique - e1: %v, %v  e2: %p, %v", e1.(*status.Error), e1, e2.(*status.Error), e2)
 	}
 }
 
-func TestFromToProto(t *testing.T) {
+func (s) TestFromToProto(t *testing.T) {
 	s := &spb.Status{
 		Code:    int32(codes.Internal),
 		Message: "test test test",
-		Details: []*apb.Any{{TypeUrl: "foo", Value: []byte{3, 2, 1}}},
+		Details: []*anypb.Any{{TypeUrl: "foo", Value: []byte{3, 2, 1}}},
 	}
 
 	err := FromProto(s)
@@ -71,7 +83,7 @@ func TestFromToProto(t *testing.T) {
 	}
 }
 
-func TestFromNilProto(t *testing.T) {
+func (s) TestFromNilProto(t *testing.T) {
 	tests := []*Status{nil, FromProto(nil)}
 	for _, s := range tests {
 		if c := s.Code(); c != codes.OK {
@@ -89,7 +101,7 @@ func TestFromNilProto(t *testing.T) {
 	}
 }
 
-func TestError(t *testing.T) {
+func (s) TestError(t *testing.T) {
 	err := Error(codes.Internal, "test description")
 	if got, want := err.Error(), "rpc error: code = Internal desc = test description"; got != want {
 		t.Fatalf("err.Error() = %q; want %q", got, want)
@@ -103,21 +115,21 @@ func TestError(t *testing.T) {
 	}
 }
 
-func TestErrorOK(t *testing.T) {
+func (s) TestErrorOK(t *testing.T) {
 	err := Error(codes.OK, "foo")
 	if err != nil {
-		t.Fatalf("Error(codes.OK, _) = %p; want nil", err.(*statusError))
+		t.Fatalf("Error(codes.OK, _) = %p; want nil", err.(*status.Error))
 	}
 }
 
-func TestErrorProtoOK(t *testing.T) {
+func (s) TestErrorProtoOK(t *testing.T) {
 	s := &spb.Status{Code: int32(codes.OK)}
 	if got := ErrorProto(s); got != nil {
 		t.Fatalf("ErrorProto(%v) = %v; want nil", s, got)
 	}
 }
 
-func TestFromError(t *testing.T) {
+func (s) TestFromError(t *testing.T) {
 	code, message := codes.Internal, "test description"
 	err := Error(code, message)
 	s, ok := FromError(err)
@@ -126,7 +138,7 @@ func TestFromError(t *testing.T) {
 	}
 }
 
-func TestFromErrorOK(t *testing.T) {
+func (s) TestFromErrorOK(t *testing.T) {
 	code, message := codes.OK, ""
 	s, ok := FromError(nil)
 	if !ok || s.Code() != code || s.Message() != message || s.Err() != nil {
@@ -137,7 +149,7 @@ func TestFromErrorOK(t *testing.T) {
 type customError struct {
 	Code    codes.Code
 	Message string
-	Details []*apb.Any
+	Details []*anypb.Any
 }
 
 func (c customError) Error() string {
@@ -145,18 +157,16 @@ func (c customError) Error() string {
 }
 
 func (c customError) GRPCStatus() *Status {
-	return &Status{
-		s: &spb.Status{
-			Code:    int32(c.Code),
-			Message: c.Message,
-			Details: c.Details,
-		},
-	}
+	return status.FromProto(&spb.Status{
+		Code:    int32(c.Code),
+		Message: c.Message,
+		Details: c.Details,
+	})
 }
 
-func TestFromErrorImplementsInterface(t *testing.T) {
+func (s) TestFromErrorImplementsInterface(t *testing.T) {
 	code, message := codes.Internal, "test description"
-	details := []*apb.Any{{
+	details := []*anypb.Any{{
 		TypeUrl: "testUrl",
 		Value:   []byte("testValue"),
 	}}
@@ -175,7 +185,7 @@ func TestFromErrorImplementsInterface(t *testing.T) {
 	}
 }
 
-func TestFromErrorUnknownError(t *testing.T) {
+func (s) TestFromErrorUnknownError(t *testing.T) {
 	code, message := codes.Unknown, "unknown error"
 	err := errors.New("unknown error")
 	s, ok := FromError(err)
@@ -184,7 +194,98 @@ func TestFromErrorUnknownError(t *testing.T) {
 	}
 }
 
-func TestConvertKnownError(t *testing.T) {
+func (s) TestFromErrorWrapped(t *testing.T) {
+	const code, message = codes.Internal, "test description"
+	err := fmt.Errorf("wrapped error: %w", Error(code, message))
+	s, ok := FromError(err)
+	if !ok || s.Code() != code || s.Message() != err.Error() || s.Err() == nil {
+		t.Fatalf("FromError(%v) = %v, %v; want <Code()=%s, Message()=%q, Err()!=nil>, true", err, s, ok, code, message)
+	}
+}
+
+type customErrorNilStatus struct {
+}
+
+func (c customErrorNilStatus) Error() string {
+	return "test"
+}
+
+func (c customErrorNilStatus) GRPCStatus() *Status {
+	return nil
+}
+
+func (s) TestFromErrorImplementsInterfaceReturnsOKStatus(t *testing.T) {
+	err := customErrorNilStatus{}
+	s, ok := FromError(err)
+	if ok || s.Code() != codes.Unknown || s.Message() != err.Error() {
+		t.Fatalf("FromError(%v) = %v, %v; want <Code()=%s, Message()=%q, Err()!=nil>, true", err, s, ok, codes.Unknown, err.Error())
+	}
+}
+
+func (s) TestFromErrorImplementsInterfaceReturnsOKStatusWrapped(t *testing.T) {
+	err := fmt.Errorf("wrapping: %w", customErrorNilStatus{})
+	s, ok := FromError(err)
+	if ok || s.Code() != codes.Unknown || s.Message() != err.Error() {
+		t.Fatalf("FromError(%v) = %v, %v; want <Code()=%s, Message()=%q, Err()!=nil>, true", err, s, ok, codes.Unknown, err.Error())
+	}
+}
+
+func (s) TestFromErrorImplementsInterfaceWrapped(t *testing.T) {
+	const code, message = codes.Internal, "test description"
+	err := fmt.Errorf("wrapped error: %w", customError{Code: code, Message: message})
+	s, ok := FromError(err)
+	if !ok || s.Code() != code || s.Message() != err.Error() || s.Err() == nil {
+		t.Fatalf("FromError(%v) = %v, %v; want <Code()=%s, Message()=%q, Err()!=nil>, true", err, s, ok, code, message)
+	}
+}
+
+func (s) TestCode(t *testing.T) {
+	const code = codes.Internal
+	err := Error(code, "test description")
+	if s := Code(err); s != code {
+		t.Fatalf("Code(%v) = %v; want <Code()=%s>", err, s, code)
+	}
+}
+
+func (s) TestCodeOK(t *testing.T) {
+	if s, code := Code(nil), codes.OK; s != code {
+		t.Fatalf("Code(%v) = %v; want <Code()=%s>", nil, s, code)
+	}
+}
+
+func (s) TestCodeImplementsInterface(t *testing.T) {
+	const code = codes.Internal
+	err := customError{Code: code, Message: "test description"}
+	if s := Code(err); s != code {
+		t.Fatalf("Code(%v) = %v; want <Code()=%s>", err, s, code)
+	}
+}
+
+func (s) TestCodeUnknownError(t *testing.T) {
+	const code = codes.Unknown
+	err := errors.New("unknown error")
+	if s := Code(err); s != code {
+		t.Fatalf("Code(%v) = %v; want <Code()=%s>", err, s, code)
+	}
+}
+
+func (s) TestCodeWrapped(t *testing.T) {
+	const code = codes.Internal
+	err := fmt.Errorf("wrapped: %w", Error(code, "test description"))
+	if s := Code(err); s != code {
+		t.Fatalf("Code(%v) = %v; want <Code()=%s>", err, s, code)
+	}
+}
+
+func (s) TestCodeImplementsInterfaceWrapped(t *testing.T) {
+	const code = codes.Internal
+	err := fmt.Errorf("wrapped: %w", customError{Code: code, Message: "test description"})
+	if s := Code(err); s != code {
+		t.Fatalf("Code(%v) = %v; want <Code()=%s>", err, s, code)
+	}
+}
+
+func (s) TestConvertKnownError(t *testing.T) {
 	code, message := codes.Internal, "test description"
 	err := Error(code, message)
 	s := Convert(err)
@@ -193,7 +294,7 @@ func TestConvertKnownError(t *testing.T) {
 	}
 }
 
-func TestConvertUnknownError(t *testing.T) {
+func (s) TestConvertUnknownError(t *testing.T) {
 	code, message := codes.Unknown, "unknown error"
 	err := errors.New("unknown error")
 	s := Convert(err)
@@ -202,10 +303,10 @@ func TestConvertUnknownError(t *testing.T) {
 	}
 }
 
-func TestStatus_ErrorDetails(t *testing.T) {
+func (s) TestStatus_ErrorDetails(t *testing.T) {
 	tests := []struct {
 		code    codes.Code
-		details []proto.Message
+		details []protoadapt.MessageV1
 	}{
 		{
 			code:    codes.NotFound,
@@ -213,7 +314,7 @@ func TestStatus_ErrorDetails(t *testing.T) {
 		},
 		{
 			code: codes.NotFound,
-			details: []proto.Message{
+			details: []protoadapt.MessageV1{
 				&epb.ResourceInfo{
 					ResourceType: "book",
 					ResourceName: "projects/1234/books/5678",
@@ -223,7 +324,7 @@ func TestStatus_ErrorDetails(t *testing.T) {
 		},
 		{
 			code: codes.Internal,
-			details: []proto.Message{
+			details: []protoadapt.MessageV1{
 				&epb.DebugInfo{
 					StackEntries: []string{
 						"first stack",
@@ -234,9 +335,9 @@ func TestStatus_ErrorDetails(t *testing.T) {
 		},
 		{
 			code: codes.Unavailable,
-			details: []proto.Message{
+			details: []protoadapt.MessageV1{
 				&epb.RetryInfo{
-					RetryDelay: &dpb.Duration{Seconds: 60},
+					RetryDelay: &durationpb.Duration{Seconds: 60},
 				},
 				&epb.ResourceInfo{
 					ResourceType: "book",
@@ -253,15 +354,18 @@ func TestStatus_ErrorDetails(t *testing.T) {
 			t.Fatalf("(%v).WithDetails(%+v) failed: %v", str(s), tc.details, err)
 		}
 		details := s.Details()
+		if len(details) != len(tc.details) {
+			t.Fatalf("len(s.Details()) = %d, want = %d.", len(details), len(tc.details))
+		}
 		for i := range details {
-			if !proto.Equal(details[i].(proto.Message), tc.details[i]) {
+			if !proto.Equal(details[i].(protoreflect.ProtoMessage), tc.details[i].(protoreflect.ProtoMessage)) {
 				t.Fatalf("(%v).Details()[%d] = %+v, want %+v", str(s), i, details[i], tc.details[i])
 			}
 		}
 	}
 }
 
-func TestStatus_WithDetails_Fail(t *testing.T) {
+func (s) TestStatus_WithDetails_Fail(t *testing.T) {
 	tests := []*Status{
 		nil,
 		FromProto(nil),
@@ -274,27 +378,27 @@ func TestStatus_WithDetails_Fail(t *testing.T) {
 	}
 }
 
-func TestStatus_ErrorDetails_Fail(t *testing.T) {
+func (s) TestStatus_ErrorDetails_Fail(t *testing.T) {
 	tests := []struct {
-		s *Status
-		i []interface{}
+		s    *Status
+		want []any
 	}{
 		{
-			nil,
-			nil,
+			s:    nil,
+			want: nil,
 		},
 		{
-			FromProto(nil),
-			nil,
+			s:    FromProto(nil),
+			want: nil,
 		},
 		{
-			New(codes.OK, ""),
-			[]interface{}{},
+			s:    New(codes.OK, ""),
+			want: []any{},
 		},
 		{
-			FromProto(&spb.Status{
+			s: FromProto(&spb.Status{
 				Code: int32(cpb.Code_CANCELLED),
-				Details: []*apb.Any{
+				Details: []*anypb.Any{
 					{
 						TypeUrl: "",
 						Value:   []byte{},
@@ -306,8 +410,8 @@ func TestStatus_ErrorDetails_Fail(t *testing.T) {
 					}),
 				},
 			}),
-			[]interface{}{
-				errors.New(`message type url "" is invalid`),
+			want: []any{
+				errors.New("invalid empty type URL"),
 				&epb.ResourceInfo{
 					ResourceType: "book",
 					ResourceName: "projects/1234/books/5678",
@@ -317,9 +421,23 @@ func TestStatus_ErrorDetails_Fail(t *testing.T) {
 		},
 	}
 	for _, tc := range tests {
-		got := tc.s.Details()
-		if !reflect.DeepEqual(got, tc.i) {
-			t.Errorf("(%v).Details() = %+v, want %+v", str(tc.s), got, tc.i)
+		details := tc.s.Details()
+		if len(details) != len(tc.want) {
+			t.Fatalf("len(s.Details()) = %d, want = %d.", len(details), len(tc.want))
+		}
+		for i, d := range details {
+			// s.Details can either contain an error or a proto message.  We
+			// want to do a compare the proto message for an Equal match, and
+			// for errors only check for presence.
+			if _, ok := d.(error); ok {
+				if (d != nil) != (tc.want[i] != nil) {
+					t.Fatalf("s.Details()[%v] was %v; want %v", i, d, tc.want[i])
+				}
+				continue
+			}
+			if !cmp.Equal(d, tc.want[i], cmp.Comparer(proto.Equal)) {
+				t.Fatalf("s.Details()[%v] was %v; want %v", i, d, tc.want[i])
+			}
 		}
 	}
 }
@@ -328,22 +446,22 @@ func str(s *Status) string {
 	if s == nil {
 		return "nil"
 	}
-	if s.s == nil {
+	if s.Proto() == nil {
 		return "<Code=OK>"
 	}
-	return fmt.Sprintf("<Code=%v, Message=%q, Details=%+v>", codes.Code(s.s.GetCode()), s.s.GetMessage(), s.s.GetDetails())
+	return fmt.Sprintf("<Code=%v, Message=%q, Details=%+v>", s.Code(), s.Message(), s.Details())
 }
 
 // mustMarshalAny converts a protobuf message to an any.
-func mustMarshalAny(msg proto.Message) *apb.Any {
-	any, err := ptypes.MarshalAny(msg)
+func mustMarshalAny(msg proto.Message) *anypb.Any {
+	m, err := anypb.New(msg)
 	if err != nil {
-		panic(fmt.Sprintf("ptypes.MarshalAny(%+v) failed: %v", msg, err))
+		panic(fmt.Sprintf("anypb.New(%+v) failed: %v", msg, err))
 	}
-	return any
+	return m
 }
 
-func TestFromContextError(t *testing.T) {
+func (s) TestFromContextError(t *testing.T) {
 	testCases := []struct {
 		in   error
 		want *Status
@@ -352,6 +470,8 @@ func TestFromContextError(t *testing.T) {
 		{in: context.DeadlineExceeded, want: New(codes.DeadlineExceeded, context.DeadlineExceeded.Error())},
 		{in: context.Canceled, want: New(codes.Canceled, context.Canceled.Error())},
 		{in: errors.New("other"), want: New(codes.Unknown, "other")},
+		{in: fmt.Errorf("wrapped: %w", context.DeadlineExceeded), want: New(codes.DeadlineExceeded, "wrapped: "+context.DeadlineExceeded.Error())},
+		{in: fmt.Errorf("wrapped: %w", context.Canceled), want: New(codes.Canceled, "wrapped: "+context.Canceled.Error())},
 	}
 	for _, tc := range testCases {
 		got := FromContextError(tc.in)
